# Developed by CodeStackers Ltd
#         
# 4 Sep 2001 - richard@codestackers.com
#              first version (0.3) completed.
###############################################################################
import Globals, AccessControl
from OFS.Folder import Folder
from BTrees.IOBTree import IOBTree
from Products.LocalFS.LocalFS import LocalFS
import time

class CategoryError(Exception):
    pass

class GroupError(Exception):
    pass

class StatusError(Exception):
    pass

class UserError(Exception):
    pass

class PriorityError(Exception):
    pass

class IssueComment(Globals.Persistent):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, comment, user_id):
        self.submit_time = int(time.time())
        self.comment = comment
        self.submitted_by = user_id

class IssueAttachment(IssueComment):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, attachment_id, comment, user_id):
        IssueComment.__init__(self, comment, user_id)
        self.attachment_id = attachment_id 

class Issue(Globals.Persistent):
    __allow_access_to_unprotected_subobjects__ = 1
    
    meta_type = 'Issue'

    def __init__(self, IssueTracker, user_id, summary, id):
        self.issuetrack = IssueTracker
        
        # initialise our BTrees
        self.comments = IOBTree()
        self.comments_length = 0
        
        self.audit_trail = IOBTree()
        self.audit_trail_length = 0
        
        self.attachments = IOBTree()
        self.attachments_length = 0
        
        self.submit_time = int(time.time())
        
        self.set_submitted_by(user_id, user_id)
        self.set_summary(user_id, summary)

        self.id = id

        self.priority = 1
        
        self.status = ''
        self.category = None
        self.group = None
        self.assigned_to = 'nobody'

    def __add_audit_trail(self, user_id, comment, before, after):
        """ Add a new event to the audit trail

        """
        a_time = int(time.time())
        # add a list to the audit trail to avoid a race condition
        # wiping out the last entry
        event = [(user_id, comment, before, after)]
        if self.audit_trail.has_key(a_time):
            self.audit_trail[a_time] += event
        else:
            self.audit_trail[a_time] = event
        self.audit_trail_length += 1

        self._p = 1

    def set_category(self, c_user_id, category):
        category = category.strip()
        if category not in self.issuetrack.categories:
            raise CategoryError, 'category "%s" does not exist' % category
        
        self.__add_audit_trail(c_user_id, 'set category',
                               self.category, category)
        self.category = category

    def set_submitted_by(self, c_user_id, submitted_by):
        # submitted_by is _not_ a string, it is wrapped in an Aquisition wrapper
        submitted_by = str(submitted_by).strip()
        if submitted_by not in self.issuetrack.visible_users:
            raise UserError, 'user "%s" is not a valid user for submitted_by' %\
                  submitted_by
        
        self.__add_audit_trail(c_user_id, 'set submitted_by',
                               getattr(self, 'submitted_by', None),
                               submitted_by)
        self.submitted_by = str(submitted_by)

    def set_group(self, c_user_id, group):
        group = group.strip()
        if group not in self.issuetrack.groups:
            raise GroupError, 'group "%s" does not exist' % group

        self.__add_audit_trail(c_user_id, 'set group', self.group, group)
        self.group = group

    def set_status(self, c_user_id, status):
        status = status.strip()
        if status not in self.issuetrack.statuses:
            raise GroupError, 'status "%s" does not exist' % status
        
        self.__add_audit_trail(c_user_id, 'set status', self.status, status)
        self.status = status
        
    def set_priority(self, c_user_id, priority):
        priority = int(priority)
        if not (priority >= 1 and priority <= self.issuetrack.max_priority):
            raise PriorityError, 'priority "%s" is not a valid priority' % \
                  priority
        
        self.__add_audit_trail(c_user_id, 'set priority',
                               self.priority, priority)
        self.priority = priority
        
    def set_summary(self, c_user_id, summary):
        summary = summary.strip()
        
        self.__add_audit_trail(c_user_id, 'set summary',
                               getattr(self, 'summary', None),
                               summary)
        self.summary = summary

    def set_assignment(self, c_user_id, assigned_to):
        assigned_to = assigned_to.strip()
        if assigned_to not in self.issuetrack.visible_users:
            raise UserError, 'user "%s" is not a valid user for assignment' % \
                  assigned_to
        
        self.__add_audit_trail(c_user_id, 'set assignment',
                               self.assigned_to, assigned_to)
        self.assigned_to = assigned_to

    def add_comment(self, user_id, comment):
        comment = comment.strip()
        
        self.__add_audit_trail(user_id, 'add comment', None, comment)
        comment_obj = IssueComment(comment, user_id)
        
        if self.comments.has_key(comment_obj.submit_time):
            self.comments[comment_obj.submit_time] += [comment_obj]
        else:
            self.comments[comment_obj.submit_time] = [comment_obj]
        self.comments_length += 1

        self._p = 1

    def add_attachment(self, user_id, filename, comment):
        filename = filename.split('\\')[-1]
        filename = filename.split('/')[-1]
        
        attachment_id = str(int(time.time()))+'_'+filename.strip()

        comment=comment.strip()

        self.__add_audit_trail(user_id, 'add attachment', None, comment)
        attachment_obj = IssueAttachment(attachment_id, comment, user_id)
        
        if self.attachments.has_key(attachment_obj.submit_time):
            self.attachments[attachment_obj.submit_time] += [attachment_obj]
        else:
            self.attachments[attachment_obj.submit_time] = [attachment_obj]
        self.attachments_length += 1
        
        self._p = 1

        return attachment_id

    def get_comments(self):
        return self.comments.values()

    def get_attachments(self):
        return self.attachments.values()

    def get_attachment(self, attachment_id):
        if self.attachments.has_key(attachment_id):
            return self.attachments[attachment_id]
        else:
            return None

    def get_number_comments(self):
        return self.comments_length

    def get_number_attachments(self):
        return self.attachments_length

class IssueTrack(Folder):
    from Products.PageTemplates.ZopePageTemplate import manage_addPageTemplate
    from Products.LocalFS.LocalFS import manage_addLocalFS    
    from Products.ZCatalog.ZCatalog import manage_addZCatalog
    from OFS.DTMLMethod import addDTMLMethod
    from OFS.Image import manage_addImage

    meta_type = 'ZShrink'

    version = 0.31

    security = AccessControl.ClassSecurityInfo()

    security.declareProtected('View management screens', 'manage_workspace',
                                                         'manage_main',
                                                         'manage')

    default_statuses = ('open', 'closed', 'deleted', 'awaiting triage',)
    default_auto_open_status = 'open'
    default_initial_status = 'awaiting triage'

    default_groups = ('feature request', 'not a bug', 'works for me', 'bug')

    default_categories = ('general',)

    default_max_priority = 9
    
    default_max_display_num = 10

    _times = (('1min', 60), ('10min', 600), ('30min', 1800), ('1h', 3600),
              ('3h', 10800), ('6h', 21600), ('12h', 43200), ('1d', 86400),
              ('2d', 172800), ('3d', 259200), ('1w', 604800), ('2w', 1209600),
              ('1m', 2419200), ('3m', 7257600), ('6m', 14515200),
              ('1y', 29030400))

    def __init__(self, id):
        self.id = id
        self.issues = IOBTree()
        self.issues_length = 0

        self.next_id = 1

    def __add_to_issues(self, key, item):
        self.issues[key] = item
        self.issues_length += 1

    def __del_from_issues(self, key):
        del(self.issues[key])
        self.issues_length -= 1

    security.declareProtected('Export issues to XML', 'export_issues_to_xml')

    def export_issues_to_xml(self, RESPONSE):
        """ Export all the issues in the database as an XML stream
        """
        import urllib

        RESPONSE.setHeader('Content-Type', 'application/xml')
        RESPONSE.write("<issues>\n")
        
        for issue in self.issues.values():
            audit_trail = issue.audit_trail
            
            RESPONSE.write(('  <issue id="%s" priority="%s" status="%s" '
                            'category="%s" group="%s" assigned_to="%s" '
                            'submit_time="%s" submitted_by="%s" '
                            'summary="%s">\n' %
                            (issue.id, issue.priority, issue.status,
                             issue.category, issue.group, issue.assigned_to,
                             issue.submit_time, issue.submitted_by,
                             urllib.quote_plus(issue.summary))
                            ))

            for comment_list in issue.comments.values():
                for comment in comment_list:
                    RESPONSE.write(('    <comment submit_time="%s" '
                                    'submitted_by="%s">\n' %
                                   (issue.submit_time, issue.submitted_by)
                                    ))
                    RESPONSE.write('%s' % urllib.quote_plus(comment.comment))
                    RESPONSE.write('    </comment>\n')
                    
            for attachment_list in issue.attachments.values():
                for attachment in attachment_list:
                    RESPONSE.write(('    <attachment submit_time="%s" '
                                    'submitted_by="%s" '
                                    'attachment_id="%s">\n' %
                                   (issue.submit_time, issue.submitted_by,
                                    issue.attachment_id)
                                    ))
                    RESPONSE.write('%s' % urllib.quote_plus(comment.comment))
                    RESPONSE.write('    </attachment>\n')
                    
            RESPONSE.write("  </issue>\n")
                
        RESPONSE.write("</issues>")
        RESPONSE.flush()

    def submit_issue(self, summary, category, group, comment=None,
                           file_attachment=None, file_description=None,
                           REQUEST={}, RESPONSE=None):
        """ Submit a new issue to the tracker.
        
        """
        user_id =getattr(REQUEST, 'AUTHENTICATED_USER', 'nobody')
        
        issue = Issue(self, user_id, summary, self.next_id)
        
        if comment:
            issue.add_comment(user_id, comment)

        if file_attachment and file_attachment.filename:
            attachment_id = issue.add_attachment(user_id,
                                                 file_attachment.filename,
                                                 file_description)
            self.file_attachments.manage_upload(file_attachment, attachment_id)

        # set the initial priority to half the maximum priority
        issue.set_priority(user_id, self.max_priority/2)

        issue.set_status(user_id, self.initial_status)

        if group:
            issue.set_group(user_id, group)

        if category:
            issue.set_category(user_id, category)

        self.__add_to_issues(self.next_id, issue)
        
        self.next_id += 1

        self.Catalog.catalog_object(issue, str(issue.id))

        if RESPONSE:
            return RESPONSE.redirect('issue_summary_screen')

    def modify_issue(self, issue_id, summary, priority,
                           category, group,
                           comment='',
                           file_attachment=None,
                           file_description=None,
                           status=None,
                           assigned_to=None,
                           REQUEST={}, RESPONSE=None):
        """ Modify an issue that already exists in the database.
        
        """
        user_id = getattr(REQUEST, 'AUTHENTICATED_USER', 'nobody')
        issue = self.get_issue(issue_id)
        
        if file_attachment and file_attachment.filename:
            attachment_id = issue.add_attachment(user_id,
                                                 file_attachment.filename,
                                                 file_description)
            self.file_attachments.manage_upload(file_attachment, attachment_id)
            
        comment = comment.strip()
        change_status = 0
        
        if comment:
            issue.add_comment(user_id, comment)

        if summary and issue.summary != summary:
            issue.set_summary(user_id, summary)

        if priority and issue.priority != int(priority):
            issue.set_priority(user_id, int(priority))

        if group and issue.group != group:
            issue.set_group(user_id, group)

        if category and issue.category != category:
            issue.set_category(user_id, category)

        if assigned_to and issue.assigned_to != assigned_to:
            issue.set_assignment(user_id, assigned_to)
            change_status = 1

        if status and issue.status != status:
            issue.set_status(user_id, status)
        elif change_status:
            if self.auto_open_status in self.statuses:
                issue.set_status(user_id, self.auto_open_status)

        self.Catalog.uncatalog_object(str(issue.id))
        self.Catalog.catalog_object(issue, str(issue.id))

        if RESPONSE:
            return RESPONSE.redirect('issue_summary_screen')
    
    def get_issues(self, REQUEST={}):
        """ Get all issues from the database, based on a given criteria.
        
        """
        sort_order = REQUEST.get('sort_order', 'ascending')
        sort_on = REQUEST.get('sort_on', 'id')
        start_val = REQUEST.get('start_val', 1)
        start_val = int(start_val)-1
        end_val = start_val + (self.max_display_num)
        
        status = REQUEST.get('status', 'open')
        if 'any' in tuple(status): status = ''
            
        submitted_by = REQUEST.get('submitted_by', 'any')
        submitted_by = type('') == type(submitted_by) and (submitted_by,) or \
                       submitted_by
        if 'any' in submitted_by: submitted_by = ''
                
        assigned_to = REQUEST.get('assigned_to', 'any')
        assigned_to = type('') == type(assigned_to) and (assigned_to,) or \
                      assigned_to
        if 'any' in assigned_to: assigned_to = ''

        group = REQUEST.get('group', 'any')
        group = type('') == type(group) and (group,) or group
        if 'any' in group: group = ''

        category = REQUEST.get('category', 'any')
        category = type('') == type(category) and (category,) or category
        if 'any' in category: category = ''
        
        id = REQUEST.get('id', '')
        if id: id = int(id)

        priority = REQUEST.get('priority', -1)
        if priority == -1: priority = ''
        else: priority = int(priority)
        
        summary = REQUEST.get('summary', '')
        if summary: summary = summary.strip().split(' ')
        sub_time = REQUEST.get('sub_time', '')
         
        if sub_time:
            for word_time, int_time in self._times:
                if word_time == sub_time:
                    sub_time = int(time.time())-int_time
                    break
                
        results = self.Catalog.searchResults({'id': id,
                                              'priority': priority,
                                              'priority_usage': 'range:min',
                                              'status': status,
                                              'submitted_by': submitted_by,
                                              'group': group,
                                              'category': category,
                                              'submit_time': sub_time,
                                              'submit_time_usage': 'range:min',
                                              'assigned_to': assigned_to,
                                              'summary': summary,
                                              'summary_usage': 'query',
                                              'sort_on': sort_on,
                                              'sort_order': sort_order})
        if not results:
            return [], 0

        length = len(results)
        results_set = results[start_val:end_val]

        return results_set, length

    def get_issue(self, issue_id):
        """ Get an issue from the database, based on its issue_id.
        
        """
        if self.issues.has_key(int(issue_id)):
            return self.issues[int(issue_id)]
        else:
            return None

    def count_issues(self):
        """ Count the number of issues in the database.

        """
        return self.issues_length

    def calc_next(self, num):
        """ Calculate the next display start point.

        """
        return int(num)+self.max_display_num

    def calc_prev(self, num):
        """ Calculate the previous display start point.

        """
        return int(num)-self.max_display_num

    def __modify_parameter(self, parameter, item_list, modify_type):
        if modify_type == 'delete': 
            for item in item_list:
                try:
                    parameter.remove(item)
                except ValueError:
                    pass
        elif modify_type == 'create':
            for item in item_list:
                parameter.append(item)
        
        parameter.sort()

        self._p = 1
 
    def url_list(self, name, alist):
        import string, urllib
        
        result = []
        for item in alist:
            result.append('%s:list=%s' % (name, urllib.quote(item)))

        return string.join(result, '&')

    security.declareProtected('Remove all issues from ZShrink',
                              'remove_all_issues')

    def remove_all_issues(self):
        """ Remove all issues from the database.
        
        """
        del(self.issues)
        self.issues = OOBTree()
        self.issues_length = 0

    def manage_addIssueTrack(self, id, file_attachment_path, RESPONSE=None):
        """ Add a IssueTrack object to a folder.
        
        """
        obj = IssueTrack(id)
        self._setObject(id, obj)
        
        issuetrack = getattr(self.aq_parent.this(), id)

        issuetrack.post_install(file_attachment_path)

        if RESPONSE:
            RESPONSE.redirect('/%s/index_html' % id)

    def manage_addIssueTrackForm(self):
       """ The form used to get the instance id from the user.
       
       """
       return """
       <html>
       <body bgcolor="#FFFFFF">
       Please type the id and base directory of the ZShrink instance:<br>
       <form name="form" action="manage_addIssueTrack"><br>
       id<br>
       <input type="text" name="id"><br>
       file attachment storage directory<br>
       <input type="text" name="file_attachment_path"><br>
       <input type="submit" value="add">
       </form>
       </body>
       </html>"""
    
    def __read_template(self, relative_path):
        import os
        
        PRODUCTS_HOME = os.path.join(Globals.SOFTWARE_HOME,
                                     'Products/ZShrink')
        TEMPLATE_PATH = os.path.join(PRODUCTS_HOME, relative_path)
        
        f = open(TEMPLATE_PATH)

        return f.read()    

    def __add_page_templates(self, templates):
        for id, title, rel_path in templates:
            self.manage_addPageTemplate(id, title,
                                        self.__read_template(rel_path))

    def __add_dtml_methods(self, methods):
        for id, title, rel_path in methods:
            self.manage_addDTMLMethod(id, title,
                                      self.__read_template(rel_path))

    def __add_images(self, images):
        import StringIO
        for id, title, rel_path in images:
            f = StringIO.StringIO(self.__read_template(rel_path))
            f.filename = id
            self.manage_addImage(id, f, title)

    def post_install(self, file_attachment_path):
        self.__initialise_product(file_attachment_path)
        
    def __initialise_product(self, file_attachment_path):
        if file_attachment_path:
            ob = LocalFS('file_attachments',
                         'File Attachments Repository',
                          file_attachment_path, None, None)
	
            self._setObject('file_attachments', ob)

        self.manage_addZCatalog('Catalog', 'Issue Tracker Catalog')

        for column in ('priority', 'submit_time', 'status', 'category',
                       'group', 'submitted_by', 'assigned_to'):
            self.Catalog.manage_addColumn(column)

        for index in ('priority', 'submit_time', 'status', 'category',
                      'group', 'submitted_by', 'assigned_to'):
            self.Catalog.manage_addIndex(index, 'FieldIndex')
            
        for index in ('summary',):
            self.Catalog.manage_addIndex(index, 'TextIndex')

        self.manage_addProperty('statuses', self.default_statuses, 'lines')
        self.manage_addProperty('auto_open_status',
                                 self.default_auto_open_status, 'string')
        self.manage_addProperty('initial_status',
                                self.default_initial_status, 'string')
        self.manage_addProperty('categories', self.default_categories, 'lines')
        self.manage_addProperty('groups', self.default_groups, 'lines')
        self.manage_addProperty('max_priority',
                                 self.default_max_priority, 'int')
        self.manage_addProperty('max_display_num',
                                 self.default_max_display_num, 'int')
        self.manage_addProperty('visible_users',
                                 self.acl_users.getUserNames(), 'lines')

        self.__add_page_templates([('layout_template',
                                    'Layout Template',
                                    'templates/default/layout_template.zpt'),
                                   ('edit_issue_screen',
                                    'Issue Editing Screen',
                                    'templates/default/edit_issue_screen.zpt'),
                                   ('issue_summary_screen',
                                    'Issue Summary Screen',
                                    'templates/default/issue_summary_screen.zpt'),
                                   ('submit_issue_screen',
                                    'Submit Issue Screen',
                                  'templates/default/submit_issue_screen.zpt'),
                                   ('index_html',
                                    'Default Screen',
                                    'templates/default/index_html.zpt'),
                                   ('navigation_bar_top',
                                    'Top Navigation Bar Template',
                                    'templates/default/navigation_bar_top.zpt'),
                                   ('navigation_bar_bottom',
                                    'Bottom Navigation Bar Template',
                                    'templates/default/navigation_bar_bottom.zpt'),
                                   ('filter_widget',
                                    'Filter Widget Template',
                                    'templates/default/filter_widget.zpt'),
                                   ('issues_widget',
                                    'Issues Widget Template',
                                    'templates/default/issues_widget.zpt'),
                                   ])
        self.__add_dtml_methods([('default.css',
                                  'Default Cascading Style Sheet',
                                  'templates/default/default.css')])

        self.__add_images([('codestackers_logo.gif',
                            'CodeStackers Logo',
                            'templates/default/codestackers_logo.gif')])
